/* VIDEOCHATSCHOOL
 * =====================================================
 * Last updated: July 11, 2007.
 * 
 * Assumes that SCHOOL is the server.
 * 
 * TODO: ATTACH PHYSICAL SLIDERS FOR BLUR AND CAMERA ANGLE. MAKE SURE VALUE MAPS TO THE APPROPRIATE VALUES.
 * 
 * New features: 
 * -------------------
 * RFID:
 * - Must be sitting on light sensor in order to check RFID. If first time sitting down, then checks rfid tag. 
 * - if user gets up, then rfid name is cleared. Next time user sits down again, rfid will be checked.
 * - if rfid is not in the database or user has no rfid, then user is assigned rfid name of "Visitor"
 * 
 * 
 * Equipment required
 * --------------------
 * Phidget Interface kit 1 with the following: 
 * Common:
 * - Light sensor on the seat of the chair for detecting presence 
 * - servo motor attached to the camera to control the camera angle
 * - proximity sensor attached to top of camera for manual way to turn on/off camera
 * - LED light plugged into the interface kit 1 output[2] for visual indication of whether you are transmitting video, transmitting awareness info, or no transmission.
 * Special:
 * - None.
 * 
 */


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace VideoChat
{
    public partial class VideoChatSchool : Form
    {
        // constants
        const int AWARENESS_INTERVAL = 3000;        
        const int PROXIMITY_MIN_TIMES = 5;
        const int BLINKING_LIGHT_INTERVAL = 500;
        const int CHECK_RFID_INTERVAL = 1000;
        const int CAMERA_AWAY_POSITION = 180;
        const int CAMERA_TOWARD_POSITION = 0;
        
        // variable declarations       
        bool present, twoWayTransmitting, cameraCapturing, blinkingRecordState, checkRFID;
        string visitorRFIDName;
        Timer awarenessTimer;
        Timer blinkingLightTimer;        
        int proxCount;
        EasyImages.CameraClient wc;
        GroupLab.GTNetworking.SharedDictionary SD;        
        GroupLab.GTNetworking.Subscription imageSubscription;
        GroupLab.GTNetworking.Subscription pingSubscription;        
        GroupLab.GTNetworking.Subscription isHereSubscription;
        GroupLab.GTNetworking.Subscription callerIDSubscription;        
                       
        /* Thread handling: Need this block of code to make sure code runs from the main windows form, and not any other thread */
        public delegate void SetControlImage(PictureBox icontrol, Bitmap image); // delegate: pointer to a function
        public void RecievedImageToControl(PictureBox icontrol, Image image)
        {
            /* Control.InvokeRequired -- gets a value indicating whether the caller must call an invoke method when making method 
             * calls to the control because the caller is on a different thread than the one the control was created on. 
             * Control.BeginInvoke() -- executes a delegate asynchronously on the thread that the control's underlying handle was created on.
             */             
            if (icontrol.InvokeRequired) 
            {
                SetControlImage theDelegate = new SetControlImage(RecievedImageToControl);
                this.BeginInvoke(theDelegate, new object[] { icontrol, image });                
            }
            else
                icontrol.Image = image;
        }//RecievedImageToControl

        public VideoChatSchool()
        {
            
            Control.CheckForIllegalCrossThreadCalls = false;
                        
            InitializeComponent();
            
            // Set up timers
            awarenessTimer = new Timer();       // awareness Timer: sends frames using the specified interval so as to keep basic awareness
            awarenessTimer.Interval = AWARENESS_INTERVAL;
            blinkingLightTimer = new Timer();   // blinking light Timer: blinks if awareness state is on
            blinkingLightTimer.Interval = BLINKING_LIGHT_INTERVAL;
         
            // variable initializations
            wc = new EasyImages.CameraClient("DefaultCamera");            
            wc.ReceivedFrame += new EasyImages.CameraClient.ReceivedFrameEventHandler(wc_ReceivedFrame);
            wc.Start();
            present = false;
            checkRFID = false;
            blinkingRecordState = false;
            twoWayTransmitting = false;
            cameraCapturing = true;
            proxCount = 0;
            visitorRFIDName = "";            

            // form settings
            this.playButton.Enabled = false;
            this.stopButton.Enabled = true;
            
            // shared dictionary and subscription declarations
            SD = new GroupLab.GTNetworking.SharedDictionary();
            pingSubscription = new GroupLab.GTNetworking.Subscription();
            imageSubscription = new GroupLab.GTNetworking.Subscription();          
            isHereSubscription = new GroupLab.GTNetworking.Subscription();
            callerIDSubscription = new GroupLab.GTNetworking.Subscription();

            // Automatically connect to shared dictionary using localhost as the server (school: server)
            openConnection();

        }//VideoChatSchool (form)

        private void awarenessTimer_Tick(object sender, EventArgs e)
        {
            if (this.twoWayTransmitting) return; //if we are transmitting video two ways b/c both people are there, we don't want to use the awarenessTimer (send at normal frame rate)
            
            if (this.SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {       
               if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server) // Server (SCHOOL)
               { 
                    // Get normal frame to send over network using videoDisplay
                    this.SD["/ServerImage"] = new EasyImages.JPEG(wc.GetFrame(), 39);                    

                    // Display what is actually sent over the network
                    if (this.SD.Contains("/ServerImage"))
                        RecievedImageToControl(this.actualDisplay, (this.SD["/ServerImage"] as EasyImages.JPEG).GetImage()); // Display what is actually sent over network
                }
            }//if

        }//awarenessTimer_Tick

        private void blinkingLightTimer_Tick(object sender, EventArgs e)
        {
            if (blinkingRecordState)
            {
                if (this.interfaceKit1 != null && this.interfaceKit1.Attached)
                    this.interfaceKit1.Outputs[2].State = !this.interfaceKit1.Outputs[2].State;
            }
        }//blinkingLightTimer_Tick

        private void wc_ReceivedFrame(object sender, EasyImages.CameraEventArgs e)
        {
            RecievedImageToControl(this.mirrorDisplay, e.Bitmap); //this.MirrorDisplay.Image = e.Bitmap;
            
        }//client_ReceivedFrame

        private void openConnection()
        {
            this.SD.Url = "tcp://localhost:TwinnedMediaSpace";
            this.SD.Opened += new EventHandler(SD_Opened);
            this.SD.Closed += new EventHandler(SD_Closed);
            this.SD.Open();                         

            // Sensor changes
            this.interfaceKit1.SensorChange += new GroupLab.SharedPhidgets.IndexSensorChangeEventHandler(this.interfaceKit1_SensorChange);

            // Timers
            awarenessTimer.Tick += new EventHandler(awarenessTimer_Tick);
            awarenessTimer.Start();
            blinkingLightTimer.Tick += new EventHandler(blinkingLightTimer_Tick);
            blinkingLightTimer.Start();
            
        }//openConnection

        private void SD_Opened(object sender, EventArgs e)
        {
            // Connects the shared dictionary to the SQL server 
            // after this, SD is connected to the server and ready to use the persistent features
            SD.ConnectToSQLServer("server=SAULNU\\SQLEXPRESS;" +
                                       "Trusted_Connection=yes;" +
                                       "database=master;" +
                                       "connection timeout=30");
            // following line loads stored values back into the shared dictionary
            SD.LoadStoredValues();


            // imageSubscription
            this.imageSubscription.BeginInit();
            this.imageSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                this.imageSubscription.Pattern = "/ClientImage";
            this.imageSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(imageSubscription_Notified);
            this.imageSubscription.EndInit();

            // pingSubscription
            this.pingSubscription.BeginInit();
            this.pingSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                this.pingSubscription.Pattern = "/ClientPing";            
            this.pingSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(pingSubscription_Notified);
            this.pingSubscription.EndInit();

            // send the first frame
            //takeSnap();

            // isHereSubscription
            this.isHereSubscription.BeginInit();
            this.isHereSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)                
            {
                this.isHereSubscription.Pattern = "/ClientIsHere";
                this.SD["/ClientIsHere"] = false;
                this.SD["/ServerIsHere"] = false;
            }
            this.isHereSubscription.Notified +=new GroupLab.GTNetworking.SubscriptionEventHandler(isHereSubscription_Notified);
            this.isHereSubscription.EndInit();

            // callerIDSubscription
            this.callerIDSubscription.BeginInit();
            this.callerIDSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                this.callerIDSubscription.Pattern = "/ClientCallerID";
            this.callerIDSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(callerIDSubscription_Notified);
            this.callerIDSubscription.EndInit();

        }// SD_Opened

    
        private void SD_Closed(object sender, System.EventArgs args)
        {            
            if (this.interfaceKit1.Attached)
                this.interfaceKit1.Outputs[2].State = false; // recording light            

        }//SD_Closed

        private void takeSnap()
        {
            if (this.SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {
                if (cameraCapturing)
                {
                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                    {
                        this.SD["/ServerImage"] = new EasyImages.JPEG(wc.GetFrame(), 39);
                    }
                }//if
            }//if
        }//takeSnap

        private void imageSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            EasyImages.JPEG j = e.Value as EasyImages.JPEG;
            if (j != null)
            {               
                RecievedImageToControl(this.videoDisplay, j.GetImage());  //this.VideoDisplay.Image = j.GetImage();                    
            }
           
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
            {
                // Display what is actually sent over network
                 if (this.SD.Contains("/ServerImage"))                    
                    RecievedImageToControl(this.actualDisplay, (this.SD["/ServerImage"] as EasyImages.JPEG).GetImage());

                // indicate that image has been recieved and can now get the next one (stop and wait)
                this.SD["/ServerPing"] = "ping!";
            }
        }//imageSubscription_Notified

        private void pingSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            if (twoWayTransmitting) takeSnap();
            
        }//pingSubscription_Notified

        private void callerIDSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            //throw new Exception("The method or operation is not implemented.");
        }//callerIDSubscription_Notified
                  
        private void checkPresence()
        {
            bool otherPresent = false; 

            if (SD.ConnectionType.Equals(GroupLab.GTNetworking.ConnectionType.Server))
                otherPresent = (bool)this.SD["/ClientIsHere"];  
                        
            // case 1: This is here, other isn't
            if (present && !otherPresent)
            {
                Console.WriteLine("State 1: this here, other isn't");                
                this.SD["/WaitingOnHome"] = true;            
                twoWayTransmitting = false;
                setRecordingStateLights("blink");
            }
            // case 2: This isn't here, other is  
            else if (!present && otherPresent)
            {
                Console.WriteLine("State 2: This not here, other is here");                
                twoWayTransmitting = false;
                setRecordingStateLights("blink");
            }
            // case 3: both there
            else if (present && otherPresent)
            {
                Console.WriteLine("State 3: Both here");                
                twoWayTransmitting = true;               
                takeSnap();
                setRecordingStateLights("solid");
            }
            // case 4: both gone
            else
            {
                Console.WriteLine("State 4: Both not here");                
                twoWayTransmitting = false;
                setRecordingStateLights("blink");
            }
        }//checkPresence

        private void isHereSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            checkPresence();
        }//isHereSubscription_Notified

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // if user chooses to close the form manually, do so in an elegant fashion (stop camera and close dictionary)
            wc.Stop();
            if (SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {
                SD.Close();
            }
        }//Form1_FormClosing
       
        private void interfaceKit1_SensorChange(object sender, GroupLab.SharedPhidgets.IndexSensorChangeEventArgs e)
        {
            if (SD.Status != GroupLab.GTNetworking.ConnectionStatus.Opened) return;

            // Light sensor on chair to detect presence
            if (e.Index == 0)
            {
                if (e.Value < 100)              // sitting in chair
                {
                    present = true;

                    if (visitorRFIDName.Length == 0)  // first time visitor has sat down, RFID name not known yet 
                    {
                        visitorRFIDName = "Visitor"; // set to visitor by default first
                        this.visitorLabel.Text = visitorRFIDName;
                        this.SD["/ServerCallerID"] = visitorRFIDName;
                        checkRFID = true; // then check RFID 
                    }
                    else if (!visitorRFIDName.Equals("Visitor") && visitorRFIDName.Length > 0) // RFID name known, no need to check anymore
                    {
                        checkRFID = false;                        
                    }
                    else if (visitorRFIDName.Equals("Visitor")) 
                    {
                        checkRFID = true;
                    }
                                        
                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                    {
                        this.SD["/ServerIsHere"] = true;
                    }
                }
                else if (e.Value >= 100)        // not sitting in chair
                {
                    present = false;

                    // if user has gotten up, clear the name (if there was a name), and do not check                    
                    visitorRFIDName = "";
                    this.SD["/ServerCallerID"] = visitorRFIDName;
                    this.visitorLabel.Text = visitorRFIDName;
                    checkRFID = false;

                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                    {
                        this.SD["/ServerIsHere"] = false;
                        this.SD["/WaitingOnHome"] = false;
                    }
                }
                checkPresence();
            }

            // Proximity sensor attached to the camera to turn camera on/off
            if (e.Index == 1)   
            {
                if (e.Value >= 200 && e.Value <= 250)       // turn camera on or off if proximity within this value
                {
                    proxCount++;
                    if (proxCount >= PROXIMITY_MIN_TIMES)
                    {
                        proxCount = 0;
                        if (this.cameraCapturing)
                            setCamera(0);
                        else 
                            setCamera(1);
                    }                    
                }//if
                else
                    proxCount = 0;
            }//if proximity sensor  
        }//interfaceKit1_SensorChange

        private void setRecordingStateLights(string state)
        {
            switch (state)
            {
                case "solid":
                    Console.WriteLine("Solid both transmitting");
                    blinkingRecordState = false;
                    if (this.interfaceKit1 != null && this.interfaceKit1.Attached)
                        this.interfaceKit1.Outputs[2].State = true; 
                    break;
                case "blink":
                    Console.WriteLine("Blinking awareness state");
                    blinkingRecordState = true;                   
                    break;
                case "none":
                    Console.WriteLine("not capturing anything");                    
                    blinkingRecordState = false;
                    if (this.interfaceKit1 != null && this.interfaceKit1.Attached)
                        this.interfaceKit1.Outputs[2].State = false;
                    break;
                default:
                    Console.WriteLine("default recording state lights");
                    break;

            }//switch            

        }//setRecordingStateLights

        private void playButton_Click(object sender, EventArgs e)
        {
            setCamera(1); // start camera                            
        }//playButton_Click

        private void stopButton_Click(object sender, EventArgs e)
        {
            setCamera(0); // stop camera
        }//stopButton_Click
      
        private void setCamera(int action)
        {
            switch (action)
            {
                case 0:     // Stop camera                    
                    wc.Stop();

                    this.playButton.Enabled = true;
                    this.stopButton.Enabled = false;

                    // use servo to change the camera angle away from person
                    if (this.servo1.Attached)
                        this.servo1.Motors[0].Position = CAMERA_AWAY_POSITION;
                    else
                        Console.WriteLine("Servo motor NOT attached");

                    awarenessTimer.Stop();                                        
                    this.twoWayTransmitting = false; // set to false because in ping_Notified, wont' be calling takeSnap to get the next frame (only send the stopped image)
                    this.cameraCapturing = false;

                    // if camera is stopped, set the recording light state
                    setRecordingStateLights("none");

                    // show the stopped image in the actual display
                    Image i = Image.FromFile("stopped.jpg");
                    RecievedImageToControl(this.actualDisplay, i);

                    // show stopped image in video display (depending on the connection type)
                    if (this.SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
                    {
                        if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                            this.SD["/ServerImage"] = new EasyImages.JPEG(i, 39);
                    }
                    break;

                case 1:     // Start camera
                    wc.Start();

                    this.playButton.Enabled = false;
                    this.stopButton.Enabled = true;

                    // use servo to change the camera angle to face towards person
                    if (this.servo1.Attached)
                        this.servo1.Motors[0].Position = CAMERA_TOWARD_POSITION;
                    else
                        Console.WriteLine("Servo motor NOT attached");

                    awarenessTimer.Start();
                    this.cameraCapturing = true;
                    this.awarenessTimer_Tick(null, null);                    
                    checkPresence();
                    break;

                default:
                    Console.WriteLine("setCamera default switch statement");
                    break;
            }//switch
        }//setCamera

        private void rfid1_Tag(object sender, GroupLab.SharedPhidgets.RFIDTagEventArgs e)
        {
            if (checkRFID)
            {
                Console.WriteLine("Checking RFID");
                if (SD.Contains("/TwinMediaRFIDS/" + e.Tag))
                {
                    visitorRFIDName = SD["/TwinMediaRFIDS/" + e.Tag].ToString();                
                }
                else
                {
                    visitorRFIDName = "Visitor";   // case where user has an RFID tag, but this tag does not exist in the database yet.
                }
                this.visitorLabel.Text = visitorRFIDName;
                this.SD["/ServerCallerID"] = visitorRFIDName;
            }           

        }//rfid1_Tag
       
     
    }// partial class Form1
}// namespace VideoChat