/* VIDEOCHATHOME
 * =====================================================
 * Last updated: July 11, 2007.
 * 
 * Assumes that HOME is the client.
 * 
 * TODO: ATTACH PHYSICAL SLIDERS FOR BLUR AND CAMERA ANGLE. MAKE SURE VALUE MAPS TO THE APPROPRIATE VALUES.
 * 
 * New features:
 * -------------------
 * TextLCD:
 * - relays a message from school to home identifying who is at school
 * - if home is not there, and has manually set blur to be a higher amount than the default value, then use the user-set value instead. (more privacy)
 * 
 * Equipment required:
 * --------------------
 * Phidget Interface kit 1 with the following: 
 * Common:
 * 
 * 
 * 
 * - servo motor attached to the camera to control the camera angle (towards user when recording, away from user when stopped recording)
 * - proximity sensor attached to top of camera for manual way to turn on/off camera
 * - LED light plugged into the interface kit 1 output[2] for visual indication of whether you are transmitting video, transmitting awareness info, or no transmission.
 * Special:
 * - LED light plugged into the interface kit 1 output[0] for indicating to HOME that someone is waiting at SCHOOL, (also plays a sound as long as person at school is 
 * sitting in the chair waiting)
 * - textCLD for relaying who is at school, waiting for home.
 * 
 */


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace VideoChatHome
{
    public partial class VideoChatHome : Form
    {
        // constants
        const int AWARENESS_INTERVAL = 3000;
        const int SOUND_INTERVAL = 1000;
        const int BLINKING_LIGHT_INTERVAL = 500;
        const int PROXIMITY_MIN_TIMES = 5;
        const int CAMERA_AWAY_POSITION = 180;
        const int CAMERA_TOWARD_POSITION = 0;
        const int WIDTH_FORM = 385;
        const int HEIGHT_TOOLS = 596;
        const int HEIGHT_NOTOOLS = 481;
        const int HOME_BLUR_AMOUNT = 5;        

        // variable declarations       
        bool present, twoWayTransmitting, homeBlurred, schoolWaiting, cameraCapturing, blinkingRecordState;
        int proxCount, blurAmount, cameraAngle;
        EasyImages.CameraClient wc;
        GroupLab.GTNetworking.SharedDictionary SD;        
        GroupLab.GTNetworking.Subscription imageSubscription;
        GroupLab.GTNetworking.Subscription pingSubscription;       
        GroupLab.GTNetworking.Subscription isHereSubscription;
        GroupLab.GTNetworking.Subscription waitingSubscription;
        GroupLab.GTNetworking.Subscription callerIDSubscription;  
        Timer awarenessTimer;
        Timer meepmeepTimer;
        Timer blinkingLightTimer;
                       
        /* Thread handling: Need this block of code to make sure code runs from the main windows form, and not any other thread */
        public delegate void SetControlImage(PictureBox icontrol, Bitmap image); // delegate: pointer to a function
        public void RecievedImageToControl(PictureBox icontrol, Image image)
        {
            /* Control.InvokeRequired -- gets a value indicating whether the caller must call an invoke method when making method 
             * calls to the control because the caller is on a different thread than the one the control was created on. 
             * Control.BeginInvoke() -- executes a delegate asynchronously on the thread that the control's underlying handle was created on.
             */             
            if (icontrol.InvokeRequired) 
            {
                SetControlImage theDelegate = new SetControlImage(RecievedImageToControl);
                this.BeginInvoke(theDelegate, new object[] { icontrol, image });                
            }
            else
                icontrol.Image = image;
        }//RecievedImageToControl

        public VideoChatHome()
        {
            Control.CheckForIllegalCrossThreadCalls = false;
            InitializeComponent();

            // Set up timers
            meepmeepTimer = new Timer();
            meepmeepTimer.Interval = SOUND_INTERVAL;
            awarenessTimer = new Timer();
            awarenessTimer.Interval = AWARENESS_INTERVAL;
            blinkingLightTimer = new Timer();
            blinkingLightTimer.Interval = BLINKING_LIGHT_INTERVAL;
            
            // variable initializations
            wc = new EasyImages.CameraClient("DefaultCamera");            
            wc.ReceivedFrame += new EasyImages.CameraClient.ReceivedFrameEventHandler(wc_ReceivedFrame);
            wc.Start();
            present = false;
            homeBlurred = true;
            blinkingRecordState = false;
            twoWayTransmitting = false;
            schoolWaiting = false;
            cameraCapturing = true;            
            proxCount = 0;
            blurAmount = 0;
            playButton.Enabled = false;
            stopButton.Enabled = true;            
            this.toolsGroupBox.Hide();
            this.Size = new System.Drawing.Size(WIDTH_FORM, HEIGHT_NOTOOLS);
           
            // shared dictionary and subscription declarations
            SD = new GroupLab.GTNetworking.SharedDictionary();
            pingSubscription = new GroupLab.GTNetworking.Subscription();
            imageSubscription = new GroupLab.GTNetworking.Subscription();            
            isHereSubscription = new GroupLab.GTNetworking.Subscription();
            waitingSubscription = new GroupLab.GTNetworking.Subscription();
            callerIDSubscription = new GroupLab.GTNetworking.Subscription();
            
        }//Form1

        private void awarenessTimer_Tick(object sender, EventArgs e)
        {
            if (this.twoWayTransmitting) return; //if we are transmitting video two ways b/c both people are there, we don't want to use the awarenessTimer (send at normal frame rate)
            
            if (this.SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {
                if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client) // Client (HOME)
                {
                    if (homeBlurred)
                    {
                        if (blurTrackBar.Value > HOME_BLUR_AMOUNT) // if what user set manually is higher than the default home blur amount, then use user-set-value instead. (more privacy)
                            blurAmount = blurTrackBar.Value;
                        else
                        {
                            blurAmount = HOME_BLUR_AMOUNT;
                            blurTrackBar.Value = blurAmount;                            
                        }
                        blurLabel.Text = "Blur Value: " + blurTrackBar.Value.ToString();                             
                    }
                    blurImage();
                }             
            }//if

        }//Form1

        private void wc_ReceivedFrame(object sender, EasyImages.CameraEventArgs e)
        {
            RecievedImageToControl(this.mirrorDisplay, e.Bitmap); //this.MirrorDisplay.Image = e.Bitmap;
            
        }//client_ReceivedFrame

        private void ConnectionButton_Click(object sender, EventArgs e)
        {
            if (this.connectionButton.Text == "Connect!")
            {               
                string url = "tcp://136.159.7.33:TwinnedMediaSpace";     // Connect to Saul's school office IP
                if (DialogResult.OK == GroupLab.GTNetworking.UrlPrompt.Show(ref url))
                {
                    this.SD.Url = url;                    
                    this.SD.Opened += new EventHandler(SD_Opened);
                    this.SD.Closed +=  new EventHandler(SD_Closed);
                    this.SD.Open(); // start client
                    this.connectionButton.Text = "Disconnect!";                    
                    this.stopButton.Enabled = true;
                    this.playButton.Enabled = false;
                    
                    // sensor changes
                    this.interfaceKit1.SensorChange += new GroupLab.SharedPhidgets.IndexSensorChangeEventHandler(this.interfaceKit1_SensorChange);
                    
                    // timers
                    awarenessTimer.Tick += new EventHandler(awarenessTimer_Tick);
                    awarenessTimer.Start();                    
                    
                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                    {
                        meepmeepTimer.Tick += new EventHandler(meepmeepTimer_Tick);
                        meepmeepTimer.Start();
                    }

                    blinkingLightTimer.Tick += new EventHandler(blinkingLightTimer_Tick);
                    blinkingLightTimer.Start();
                }
            }
            else
            {
                this.connectionButton.Text = "Connect!";
                this.SD.Close(); // close the connection with the shared dictionary
                this.playButton.Enabled = true;
                this.stopButton.Enabled = false;
            }
        }//ConnectionButton_Click

        private void blinkingLightTimer_Tick(object sender, EventArgs e)
        {
            if (blinkingRecordState)
            {
                this.interfaceKit1.Outputs[2].State = !this.interfaceKit1.Outputs[2].State;
            }
        }//blinkingLightTimer_Tick

        private void meepmeepTimer_Tick(object sender, EventArgs e)
        {
            if (schoolWaiting && !present)
            {
                this.interfaceKit1.Outputs[0].State = !this.interfaceKit1.Outputs[0].State;
                PlayWav.Play("meepmeep.wav");
            }
        }//meepmeepTimer_Tick

        private void SD_Opened(object sender, EventArgs e)
        {
            // imageSubscription
            this.imageSubscription.BeginInit();
            this.imageSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.imageSubscription.Pattern = "/ServerImage";
            this.imageSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(imageSubscription_Notified);
            this.imageSubscription.EndInit();

            // pingSubscription
            this.pingSubscription.BeginInit();
            this.pingSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.pingSubscription.Pattern = "/ServerPing";
            this.pingSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(pingSubscription_Notified);
            this.pingSubscription.EndInit();

            // send the first frame
            //takeSnap();
        
            // isHereSubscription
            this.isHereSubscription.BeginInit();
            this.isHereSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.isHereSubscription.Pattern = "/ServerIsHere";            
            this.isHereSubscription.Notified +=new GroupLab.GTNetworking.SubscriptionEventHandler(isHereSubscription_Notified);
            this.isHereSubscription.EndInit();
            
            // waitingSubscription
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client) // Client (HOME)
            {
                this.waitingSubscription.BeginInit();
                this.waitingSubscription.Dictionary = this.SD;
                this.waitingSubscription.Pattern = "/WaitingOnHome"; // (subscribe to the opposite pattern - school)
                this.waitingSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(waitingSubscription_Notified);
                this.waitingSubscription.EndInit();
            }

            // callerIDSubscription
            this.callerIDSubscription.BeginInit();
            this.callerIDSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.callerIDSubscription.Pattern = "/ServerCallerID";
            this.callerIDSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(callerIDSubscription_Notified);
            this.callerIDSubscription.EndInit();

        }// SD_Opened


        private void SD_Closed(object sender, System.EventArgs args)
        {
            // turn off all phidget states
            if (this.interfaceKit1.Attached)
            {
                this.interfaceKit1.Outputs[0].State = false; // someone waiting at school
                this.interfaceKit1.Outputs[2].State = false; // recording light                            
            }
            if (this.textLCD1.Attached)
            {
                this.textLCD1.Display = "";
            }

        }//SD_Closed

        private void takeSnap()
        {
            if (this.SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {
                if (cameraCapturing)
                {
                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                    {
                        this.SD["/ClientImage"] = new EasyImages.JPEG(wc.GetFrame(), 39);
                    }
                }//if
            }//if
        }//takeSnap

        private void imageSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            EasyImages.JPEG j = e.Value as EasyImages.JPEG;
            if (j != null)
            {               
                RecievedImageToControl(this.videoDisplay, j.GetImage());  //this.VideoDisplay.Image = j.GetImage();                    
            }
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
            {
                // Display what is actually sent over network
                if (this.SD.Contains("/ClientImage"))
                   RecievedImageToControl(this.actualDisplay, (this.SD["/ClientImage"] as EasyImages.JPEG).GetImage());         
           
                // indicate that image has been recieved and can now get the next one (stop and wait)
                this.SD["/ClientPing"] = "ping!"; 
            }            
        }//imageSubscription_Notified

        private void pingSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            if (twoWayTransmitting) takeSnap();
            
        }//pingSubscription_Notified

        private void waitingSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            if ((bool)e.Value)
            {
                schoolWaiting = true;  // Alert home that someone is waiting to chat at home
            }
            else
            {
                schoolWaiting = false;
                this.interfaceKit1.Outputs[0].State = false;
            }                        
        }//waitingSubscription_Notified

        private void callerIDSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            if (this.SD.Contains("/ServerCallerID"))
            {
                this.visitorLabel.Text = this.SD["/ServerCallerID"].ToString();
                this.textLCD1.Display = "You are talking to: " + this.SD["/ServerCallerID"].ToString();      
            }


        }// SD_Opened

     
        private void checkPresence()
        {
            bool otherPresent = false;
            if (SD.ConnectionType.Equals(GroupLab.GTNetworking.ConnectionType.Client))
                otherPresent = (bool)this.SD["/ServerIsHere"];

            // case 1: This is here, other isn't
            if (present && !otherPresent)
            {
                Console.WriteLine("State 1: this here, other isn't");
                homeBlurred = false;                
                twoWayTransmitting = false;
                setRecordingStateLights("blink");
            }
            // case 2: This isn't here, other is  
            else if (!present && otherPresent)
            {
                Console.WriteLine("State 2: This not here, other is here");
                homeBlurred = true;
                twoWayTransmitting = false;
                setRecordingStateLights("blink");
            }
            // case 3: both there
            else if (present && otherPresent)
            {
                Console.WriteLine("State 3: Both here");
                homeBlurred = false;
                twoWayTransmitting = true;               
                takeSnap();
                setRecordingStateLights("solid");
            }
            // case 4: both gone
            else
            {
                Console.WriteLine("State 4: Both not here");
                homeBlurred = true;
                twoWayTransmitting = false;
                setRecordingStateLights("blink");
            }
        }//checkPresences

        private void isHereSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            checkPresence();
        }//isHereSubscription_Notified

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // if user chooses to close the form manually, do so in an elegant fashion (stop camera and close dictionary)
            wc.Stop();
            setRecordingStateLights("none");
            if (SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {
                SD.Close();
            }
        }//Form1_FormClosing
        
        private void interfaceKit1_SensorChange(object sender, GroupLab.SharedPhidgets.IndexSensorChangeEventArgs e)
        {
            if (SD.Status != GroupLab.GTNetworking.ConnectionStatus.Opened) return;

            // Light sensor on chair to detect presence
            if (e.Index == 0)
            {
                if (e.Value < 100)
                {
                    present = true;
                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                    {
                        this.SD["/ClientIsHere"] = true;
                    }
                }
                else if (e.Value >= 100)
                {
                    present = false;

                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                    {
                        this.SD["/ClientIsHere"] = false;
                        this.SD["/WaitingOnHome"] = false;
                    }
                }
                checkPresence();
            }

            // Proximity sensor attached to the camera to turn camera on/off
            else if (e.Index == 1)
            {
                if (e.Value >= 200 && e.Value <= 250)       // turn camera on or off if proximity within this value
                {
                    proxCount++;
                    if (proxCount >= PROXIMITY_MIN_TIMES)
                    {
                        proxCount = 0;
                        if (this.cameraCapturing)
                            setCamera(0);
                        else
                            setCamera(1);
                    }
                }//if
                else
                    proxCount = 0;
            }//if proximity sensor

            // Camera angle slider
            else if (e.Index == 2)   // TODO: THESE VALUES NEED TO BE MORE PRECISE (MAPPING NEEDS TO BE FIXED) -- HOW MANY TICKS ON TRACKBAR, AND HOW MANY TICKS 
            {
                if (e.Value >= 0 && e.Value <= 400)
                    cameraAngle = 0;
                else if (e.Value > 400 && e.Value <= 800)
                    cameraAngle = 90;
                else if (e.Value > 800 && e.Value <= 1000)
                    cameraAngle = 180;

                // use physical slider value to set graphical trackbar
                this.cameraAngleTrackBar.Value = cameraAngle;
                this.cameraAngleLabel.Text = "Camera Angle: " + this.cameraAngleTrackBar.Value.ToString();

                setCameraAngle();
            }//if slider (camera angle)

            // Blur slider
            else if (e.Index == 3)   //TODO: these values need to fit with the slider
            {                
                if (e.Value >= 0 && e.Value <= 10)
                    blurAmount = 0;
                else if (e.Value > 10 && e.Value <= 150)
                    blurAmount = 1;
                else if (e.Value > 150 && e.Value <= 300)
                    blurAmount = 2;
                else if (e.Value > 300 && e.Value <= 450)
                    blurAmount = 3;
                else if (e.Value > 450 && e.Value <= 600)
                    blurAmount = 4;
                else if (e.Value > 600 && e.Value <= 750)
                    blurAmount = 5;
                else if (e.Value > 750 && e.Value <= 900)
                    blurAmount = 6;
                else if (e.Value > 900 && e.Value <= 1000)
                    blurAmount = 7;

                // use physical slider value to set graphical trackbar
                this.blurTrackBar.Value = blurAmount;
                this.blurLabel.Text = "Blur Value: " + this.blurTrackBar.Value.ToString();

                blurImage();

            }//if slider (blur)


        }//interfaceKit1_SensorChange

        private void blurImage()
        {
            if (!cameraCapturing) return;
            
            // get current frame
            Bitmap bmp = new Bitmap(wc.GetFrame());
            
            switch (blurAmount)
            {
                case 0:
                    // do nothing, no blur.
                    break;
                case 1:
                    EasyImages.ImageManipulator.Distort(EasyImages.DistortionStyle.Blur1, bmp);                        
                    break;
                case 2:
                    EasyImages.ImageManipulator.Distort(EasyImages.DistortionStyle.Blur2, bmp);
                    break;
                case 3:
                    EasyImages.ImageManipulator.Distort(EasyImages.DistortionStyle.Blur3, bmp);
                    break;
                case 4:
                    EasyImages.ImageManipulator.Distort(EasyImages.DistortionStyle.Blur4, bmp);
                    break;
                case 5:
                    EasyImages.ImageManipulator.Distort(EasyImages.DistortionStyle.Blur5, bmp);
                    break;
                case 6:
                    EasyImages.ImageManipulator.Distort(EasyImages.DistortionStyle.Blur6, bmp);
                    break;
                case 7:
                    EasyImages.ImageManipulator.Distort(EasyImages.DistortionStyle.Blur7, bmp);
                    break;
                default:
                    Console.WriteLine("Default distortion type!");
                    break;
            }//switch  

            
            
            // store blurred image into pattern for ImageSubscription
            this.SD["/ClientImage"] = new EasyImages.JPEG(bmp, 39);

            // Display what is actually sent over network in actualDisplay
            if (this.SD.Contains("/ClientImage"))
                RecievedImageToControl(this.actualDisplay, (this.SD["/ClientImage"] as EasyImages.JPEG).GetImage()); 
        }//blurImage

        private void setCameraAngle()
        {
            if (this.servo1.Attached)
                this.servo1.Motors[0].Position = cameraAngle;                
            else
                Console.WriteLine("servo1 NOT attached");
        }//setCameraAngle

        private void setRecordingStateLights(string state)
        {
            switch (state)
            {
                case "solid":
                    Console.WriteLine("Solid both transmitting");
                    blinkingRecordState = false;                   
                    this.interfaceKit1.Outputs[2].State = true; 
                    break;
                case "blink":
                    Console.WriteLine("Blinking awareness state");
                    blinkingRecordState = true;                   
                    break;
                case "none":
                    Console.WriteLine("not capturing anything");                    
                    blinkingRecordState = false;
                    this.interfaceKit1.Outputs[2].State = false;
                    break;
                default:
                    Console.WriteLine("default recording state lights");
                    break;

            }//switch            

        }//setRecordingStateLights

        private void toolsButton_Click(object sender, EventArgs e)
        {
            if (this.Height == HEIGHT_TOOLS) //if big window, then shrink it
            {
                this.toolsButton.Image = Image.FromFile("plus.jpg");
                this.toolsGroupBox.Hide();
                this.Size = new System.Drawing.Size(WIDTH_FORM, HEIGHT_NOTOOLS);
            }
            else if (this.Height == HEIGHT_NOTOOLS)   // if small window, then enlarge it
            {
                this.toolsButton.Image = Image.FromFile("minus.jpg");
                this.toolsGroupBox.Show(); 
                this.Size = new System.Drawing.Size(WIDTH_FORM, HEIGHT_TOOLS);                               
            }
            
        }//toolsButton_Click

        private void playButton_Click(object sender, EventArgs e)
        {
            setCamera(1); // start camera
        }//playButton_Click

        private void stopButton_Click(object sender, EventArgs e)
        {
            setCamera(0); // stop camera
        }//stopButton_Click
        
        private void setCamera(int action)
        {
            switch (action)
            {
                case 0:     // Stop camera                    
                    wc.Stop();

                    awarenessTimer.Stop();                                      
                    this.twoWayTransmitting = false; // set to false because in ping_Notified, wont' be calling takeSnap to get the next frame (only send the stopped image)
                    this.cameraCapturing = false;
                    this.playButton.Enabled = true;
                    this.stopButton.Enabled = false;

                    // use servo to change the camera angle away from person
                    if (this.servo1.Attached)
                        this.servo1.Motors[0].Position = CAMERA_AWAY_POSITION;
                    else
                        Console.WriteLine("Servo motor NOT attached");

                    // if camera is stopped, set the recording light state
                    setRecordingStateLights("none");

                    // show the stopped image in the actual display
                    Image stoppedImage = Image.FromFile("stopped.jpg");
                    RecievedImageToControl(this.actualDisplay, stoppedImage);

                    // show stopped image in video display (depending on the connection type)
                    if (this.SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
                    {
                        if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                            this.SD["/ClientImage"] = new EasyImages.JPEG(stoppedImage, 39);
                    }
                    break;

                case 1:     // Start camera
                    wc.Start();

                    // use servo to change the camera angle to face towards person
                    if (this.servo1.Attached)
                        this.servo1.Motors[0].Position = CAMERA_TOWARD_POSITION;
                    else
                        Console.WriteLine("Servo motor NOT attached");

                    awarenessTimer.Start();                    
                    this.cameraCapturing = true;
                    this.awarenessTimer_Tick(null, null);
                    this.playButton.Enabled = false;
                    this.stopButton.Enabled = true;
                    checkPresence();
                    break;

                default:
                    Console.WriteLine("setCamera default switch statement");
                    break;
            }
        } //setCamera

        private void blurTrackBar_Scroll(object sender, EventArgs e)
        {
            blurAmount = this.blurTrackBar.Value;
            this.blurLabel.Text = "Blur Value: " + this.blurTrackBar.Value.ToString();
            blurImage();
        }//blurTrackBar_Scroll

        private void cameraAngleTrackBar_Scroll(object sender, EventArgs e)
        {
            cameraAngle = this.cameraAngleTrackBar.Value;
            this.cameraAngleLabel.Text = "Camera Angle: " + this.cameraAngleTrackBar.Value.ToString();
            setCameraAngle();

        }//cameraAngleTrackBar_Scroll

    
    }// partial class Form1
}// namespace VideoChat